"""CLI for manipulating and parsing batch data files."""

from pathlib import Path
from typing import List, Optional

import requests
import typer
from geneweaver.client.parser import batch
from geneweaver.core.parse.exceptions import InvalidBatchValueLineError
from pydantic import ValidationError

cli = typer.Typer()


@cli.command()
def to_csv(
    batch_file: Path,
    output_directory: Optional[Path] = None,
    prefix: Optional[str] = None,
    geneset_ids: Optional[List[str]] = None,
) -> None:
    """Export a batch file as a csv file.

    :param batch_file: The batch file to create the csv file from.
    :param output_directory: The output directory to write the csv file to.
                             (default: current working directory)
    :param prefix: A prefix to apply to the csv filename.
    :param geneset_ids: The geneset ids that match those in the batch file (in order).
    :return: The path to the csv file.
    """
    output_files = batch.to_csv(batch_file, output_directory, prefix, geneset_ids)

    for file in output_files:
        print(f"Created {file}")


@cli.command()
def to_csv_indexed(
    batch_file: Path, index_file: Path, output_directory: Optional[Path] = None
) -> list:
    """Export a batch file as a csv file, using the index file to name the csv files.

    (This is experimental)

    :param batch_file: The batch file to create the csv file from.
    :param index_file: The index file to use to name the csv files.
    :param output_directory: Where to write the csv files to.
                             (default: current working directory)
    :return: A list of the paths to the created csv files.
    """
    output_files = batch.to_csv_indexed(batch_file, index_file, output_directory)

    for file in output_files:
        print(f"Created {file}")

    return output_files


@cli.command()
def download_genesets(
    index_file: Path,
    session: str,
    output_directory: Optional[Path] = None,
    hash_header: bool = False,
) -> None:
    """Download genesets from the Geneweaver API.

    (experimental)
    :param index_file: The index file to use to name the csv files.
    :param session: The session cookie to use to authenticate with the Geneweaver API.
    :param output_directory: Where to write the csv files to.
    :param hash_header: Whether to hash the header of the csv files.
    :return: None
    """
    # URL of the Flask app
    base_url = "https://geneweaver.org"

    index_data = batch.read_index_file(index_file)
    genesets = {}
    output_files = []
    skipped = []
    with requests.Session() as s:
        s.cookies.set("session", session)

        for index, gsid in enumerate(index_data["GW gene set id"]):
            if gsid != "NA":
                disease = index_data["disease name"][index].lower().replace(" ", "_")
                response = s.get(f"{base_url}/exportBatch/{gsid[2:]}")
                if response.ok:
                    print(f"Found {gsid}")
                    genesets[gsid] = response.text
                    try:
                        geneset = batch.batch.process_lines(genesets[gsid])[0]
                    except (ValidationError, InvalidBatchValueLineError) as e:
                        print(f"Failed parsing {gsid}")
                        skipped.append(gsid)
                        print(e)
                    output_files.append(
                        batch.write_geneset_to_csv(
                            geneset,
                            index_data["UBERON id"][index],
                            output_directory,
                            disease,
                            gsid,
                            hash_header=hash_header,
                        )
                    )
                else:
                    print(f"Skipping {gsid}")
                    skipped.append(gsid)

    # TODO: Cocaine related sets not in the index file
    # with requests.Session() as s:
    #
    #     for gs in range(407450, 407481):
    #         if response.ok:
    #             output_files.append(batch.write_geneset_to_csv(
    #                 geneset, None, output_directory, 'cocaine', f'GS{gs}')

    for output_file in output_files:
        print(f"Created {output_file}")

    for skip in skipped:
        print(f"Skipped {skip}")



"""Authentication code for the GeneWeaver client."""

import time
from typing import Any, Dict, Optional

import jwt
import requests
from auth0.authentication.token_verifier import (
    AsymmetricSignatureVerifier,
    TokenVerifier,
)
from geneweaver.client.core import app_dir
from geneweaver.client.core.config import settings
from geneweaver.client.exceptions import AuthenticationError


def login() -> None:
    """Run the device authorization flow."""
    device_code_data = _get_device_code_data()
    _print_device_code_instructions(device_code_data)
    token_data = _poll_for_flow_completion(device_code_data)
    app_dir.save_auth_token(token_data)


def get_id_token() -> Optional[str]:
    """Get the ID token from the authentication token file.

    :returns: The ID token.
    """
    return _get_token_data_value_or_none("id_token")


def get_access_token() -> Optional[str]:
    """Get the Access token from the authentication token file.

    :returns: The ID token.
    """
    token = _get_token_data_value_or_none("access_token")
    if access_token_expired(token):
        refresh_token()
        token = _get_token_data_value_or_none("access_token")
    return token


def _get_token_data_value_or_none(token_data_key: str) -> Optional[str]:
    token_data = app_dir.get_auth_token()

    if token_data is None:
        return None

    return token_data.get(token_data_key)


def validate_token(token: str) -> None:
    """Verify the token and its precedence.

    :param token:
    """
    jwks_url = "https://{}/.well-known/jwks.json".format(settings.AUTH_DOMAIN)
    issuer = "https://{}/".format(settings.AUTH_DOMAIN)
    sv = AsymmetricSignatureVerifier(jwks_url)
    tv = TokenVerifier(
        signature_verifier=sv, issuer=issuer, audience=settings.AUTH_CLIENT_ID
    )
    tv.verify(token)


def access_token_expired(access_token: str) -> bool:
    """Check if the access token is unexpired."""
    token_data = app_dir.get_auth_token()
    try:
        jwt.decode(
            token_data["access_token"],
            algorithms=settings.AUTH_ALGORITHMS,
            options={"verify_signature": False},
        )
        return False
    except jwt.ExpiredSignatureError:
        return True


def refresh_token() -> None:
    """Refresh the access token."""
    token_data = app_dir.get_auth_token()
    refresh_token = token_data["refresh_token"]
    payload = {
        "grant_type": "refresh_token",
        "client_id": settings.AUTH_CLIENT_ID,
        "refresh_token": refresh_token,
    }
    response = requests.post(
        "https://{}/oauth/token".format(settings.AUTH_DOMAIN), data=payload
    )
    token_data = response.json()
    token_data["refresh_token"] = app_dir.get_auth_token()["refresh_token"]
    app_dir.save_auth_token(token_data)


def current_user(id_token: str) -> Dict[str, str]:
    """Get the current user from the ID token."""
    return jwt.decode(
        id_token,
        algorithms=settings.AUTH_ALGORITHMS,
        options={"verify_signature": False},
    )


def _device_code_payload() -> Dict[str, str]:
    return {
        "client_id": settings.AUTH_CLIENT_ID,
        "scope": " ".join(settings.AUTH_SCOPES),
    }


def _get_device_code_data() -> dict:
    device_code_response = requests.post(
        "https://{}/oauth/device/code".format(settings.AUTH_DOMAIN),
        data=_device_code_payload(),
    )

    if device_code_response.status_code != 200:
        raise AuthenticationError("Error generating the device code")

    return device_code_response.json()


def _print_device_code_instructions(device_code_data: Dict[str, Any]) -> None:
    print(
        "1. On your computer or mobile device navigate to: ",
        device_code_data["verification_uri_complete"],
    )
    print("2. Enter the following code: ", device_code_data["user_code"])


def _token_payload(device_code: str) -> Dict[str, str]:
    return {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code,
        "scope": "offline_access",
        "client_id": settings.AUTH_CLIENT_ID,
    }


def _poll_for_flow_completion(
    device_code_data: Dict[str, Any]
) -> Optional[Dict[str, any]]:
    authenticated = False
    token_data = None
    while not authenticated:
        token_response = requests.post(
            "https://{}/oauth/token".format(settings.AUTH_DOMAIN),
            data=_token_payload(device_code_data["device_code"]),
        )

        token_data = token_response.json()
        if token_response.status_code == 200:
            print("Authenticated!")
            print("- Id Token: {}...".format(token_data["id_token"][:10]))
            validate_token(token_data["id_token"])
            authenticated = True
        elif token_data["error"] not in ("authorization_pending", "slow_down"):
            raise AuthenticationError(token_data["error_description"])
        else:
            time.sleep(device_code_data["interval"])

    return token_data



"""The main entrypoint to the GeneWeaver CLI client."""

# ruff: noqa: B008
import pkg_resources
import typer
from geneweaver.client.cli import alpha, beta

cli = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")

cli.add_typer(alpha.cli, name="alpha", help=alpha.HELP_MESSAGE)
cli.add_typer(beta.cli, name="beta", help=beta.HELP_MESSAGE)


def version_callback(version: bool) -> None:
    """Print the version of the GeneWeaver CLI client."""
    if version:
        version = pkg_resources.get_distribution("geneweaver-client").version
        typer.echo(f"GeneWeaver CLI client (gweave) version: {version}")
        raise typer.Exit(code=0)


@cli.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=version_callback),
    quiet: bool = typer.Option(False, "--quiet", "-q", help="Suppress all output."),
    pretty: bool = typer.Option(
        False, "--pretty", "-p", help="Pretty print data output."
    ),
) -> None:
    """GeneWeaver CLI client."""
    ctx.obj = {"quiet": quiet, "pretty": pretty}



"""Root of the alpha datasets subcommand."""

import typer

HELP_MESSAGE = """
Tools and utilities for interacting with the Geneweaver Client datasets.

These datasets are available to help you get started with Geneweaver, and do not reflect
 the full collection of available datasets.
"""

cli = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")



"""Functions that warp the Geneweaver /genes endpoint."""

from typing import List, Optional

from geneweaver.client.api.utils import format_endpoint, sessionmanager
from geneweaver.core.enum import GeneIdentifier, Species

ENDPOINT = "genes"


def map_homologs(
    access_token: str,
    source_ids: List[str],
    target_id_type: GeneIdentifier,
    source_id_type: Optional[GeneIdentifier] = None,
    target_species: Optional[Species] = None,
    source_species: Optional[Species] = None,
) -> dict:
    """Map homologs between species.

    :param access_token: User access token
    :param source_ids: List of source gene IDs.
    :param target_id_type: Target gene ID type (one of GeneIdentifier).
    :param source_id_type: Source gene ID type (one of GeneIdentifier).
    :param target_species: Target species (one of Species).
    :param source_species: Source species (one of Species).
    """
    post_args = {
        "source_ids": source_ids,
        "target_gene_id_type": int(target_id_type),
    }
    if source_id_type is not None:
        post_args["source_gene_id_type"] = int(source_id_type)

    if target_species is not None:
        post_args["target_species"] = int(target_species)

    if source_species is not None:
        post_args["source_species"] = int(source_species)

    with sessionmanager(token=access_token) as session:
        resp = session.post(format_endpoint(ENDPOINT, "homologs"), json=post_args)

    return resp.json()



# noqa: E501
"""Gene expression service.

The search operations on the db
are supported using a json object. This intentionally
wraps the underlying BigQuery database for the reasons
of security and scalability.
"""
import io
from collections import OrderedDict
from dataclasses import dataclass, fields
from enum import Enum
from io import StringIO
from typing import List, Mapping, Set

import numpy
import pandas
import requests
from geneweaver.client.core.config import settings
from pandas import DataFrame
from requests.models import Response

SourceType = Enum("Source", ["IMPUTED", "EXPERIMENT"])
"""
The source either experimentally determined or imputed using
machine learning ridge regression. For strain recommender, the
imputed sources are most often used.
@usage: from org.jax.gedb import SourceType
"""

Sex = Enum("Sex", ["Female", "Male", "Both"])
"""
The imputed sex field. This is not direct biological sex
but a representation of the input sex(s) to the alg.
"""


# TODO Should data access objects go in api?
@dataclass
class DataRequest:
    """Object to request data e.g. a search.

    This request mirrors the available fields in:
    gene-expr-service/src/main/java/org/jax/jcpg/dao/DataRequest.java
    You cannot rename fields in the client without breaking the API
    These are a subset of the fields. TODO (low priority): Add all.
    """

    geneIds: List[str] = None  # noqa: N815
    strains: List[str] = None  # noqa: N815
    tissue: str = None  # noqa: N815
    sourceType: SourceType = None  # noqa: N815
    sexes: List[str] = None  # noqa: N815


# TODO Should data access objects go in api?
@dataclass
class DataResult:
    """Object which contains results.

    This result mirrors the available fields in:
    gene-expr-service/src/main/java/org/jax/jcpg/dao/DataResult.java
    You cannot rename fields in the client without breaking the API.
    These are a subset of the fields. TODO (low priority): Add all.
    """

    values: List[float] = None  # noqa: N815
    names: List[str] = None  # noqa: N815
    weights: List[float] = None  # noqa: N815
    geneIds: List[str] = None  # noqa: N815
    strains: List[str] = None  # noqa: N815
    strain: str = None  # noqa: N815
    tissue: str = None  # noqa: N815
    sexes: List[str] = None  # noqa: N815


@dataclass
class StrainResult:
    """Object which contains results for strain search."""

    gene_ids: List[str] = None  # noqa: N815
    gene_names: List[str] = None  # noqa: N815
    strain: str = None  # noqa: N815
    strain_expressions: Mapping[str, List[float]] = None  # noqa: N815


@dataclass
class NullVarianceRequest:
    """Simple object for getting spearmanrho variance."""

    id: str = None  # noqa: N815
    scores: List[float] = None  # noqa: N815
    rSz: int = 1000  # noqa: N815


# TODO Should data access objects go in api?
@dataclass
class Metadata:
    """Object to contain Metadata.

    The Metadata is available for various properties by searching
    the database.

    Example:
    -------
    {
        "ingestid":"efb7dc30-f529-441c-bfc5-97876e3536ba",
        "modelversion":"ridge_v1_2_1",
        "population":"GenomeMUSter_v2",
        "tissue":"maxilla",
        "sourcetype":"IMPUTED",
        "species":"Mus musculus",
        "uberon":"0002397",
        "bucketname":"jax-compsci-nc-dev-01-gene-expression-dev",
        "dataobject":"imputed/imputed_maxilla_ridge_v1_2_1-data.tsv.gz",
        "weightobject":"imputed/imputed_maxilla_ridge_v1_2_1-weight.tsv.gz",
        "metaobject":"imputed/imputed_maxilla_ridge_v1_2_1-meta.tsv"
    }

    """

    ingestid: str = None  # noqa: N815
    modelversion: str = None  # noqa: N815
    population: str = None  # noqa: N815
    tissue: str = None  # noqa: N815
    sourcetype: SourceType = None  # noqa: N815
    species: str = None  # noqa: N815
    uberon: str = None  # noqa: N815
    bucketname: str = None  # noqa: N815
    dataobject: str = None  # noqa: N815
    weightobject: str = None  # noqa: N815
    metaobject: str = None  # noqa: N815


@dataclass
class Bulk:
    """Object to contain Bulk expression data."""

    genename: str = None  # noqa: N815
    geneid: str = None  # noqa: N815
    exprnames: List[str] = None  # noqa: N815
    exprvalues: List[float] = None  # noqa: N815
    weightvalues: List[float] = None  # noqa: N815
    ingestid: str = None  # noqa: N815


class GeneExpressionDatabaseClient:
    """Gene Expression Database Client.

    Client object to which you make DataRequests and from which
    you get a list of DataResults for your search. It is possible to construct
    searches which the server takes a very long time to construct. In general
    the examples of supported searches are listed on the swagger page of the server.
    e.g. at https://geneweaver-dev.jax.org/gedb/ under for instance
    the 'gene/expression/search' endpoint.
    """

    def __init__(self, url: str = None, auth_proxy: str = None) -> None:
        """Create a GeneExpressionDatabaseClient from a URL.

        @param url: The optional URL to which we will connect
        when making gedb server queries.
        @param auth_proxy: The optional value of a cookie to
        connect to https version of the API.
        """
        if url is None:
            url = settings.GEDB
        self.url = url
        self.auth_proxy = auth_proxy

    def search(self, drequest: DataRequest) -> List[DataResult]:
        """Do a gene expression search on the Gene Expression Database.

        using fields available in the DataRequest object.
        @param drequest: The request which we want to make to the client
        to get results.
        """
        url = self._get_search_url()

        response = self._post(url, drequest.__dict__)

        # TODO Not sure if need to deal with typing here.
        # Need to write test to check.
        return [self._class_from_args(DataResult, item) for item in response.json()]

    def search_expression(self, drequest: DataRequest) -> List[StrainResult]:
        """Do a gene expression search on the Gene Expression Database.

        using fields available in the DataRequest object.
        This call is the same as search, however it specifically
        orders the expressions by strain, individual and sex which
        is the required ordering for concordance calculation.

        @param drequest: The request which we want to make to the client
        to get results.
        """
        url = self._get_search_expression_url()

        response = self._post(url, drequest.__dict__)

        # TODO Not sure if need to deal with typing here.
        # Need to write test to check.
        return [self._class_from_args(StrainResult, item) for item in response.json()]

    def distinct(self, field: str) -> Set[str]:
        """Get list of unique fields from metadata.

        @param field: For instance to get the
         strains return field = "tissue"
        Available fields are listed in swagger
        e.g. https://geneweaver-dev.jax.org/gedb/ under
        meta/distinct/strain.
        """
        url = "{}/{}".format(self._get_distinct_url(), field)

        response = self._get(url)

        return response.json()

    def get_meta(self, tissue: str) -> List[Metadata]:
        """Get metadata from database."""
        url = "{}/{}".format(self._get_meta_url(), tissue)
        response = self._get(url)
        return [self._class_from_args(Metadata, item) for item in response.json()]

    def _class_from_args(self, class_name: object, arg_dict: dict) -> object:
        field_set = {f.name for f in fields(class_name) if f.init}
        filtered = {k: v for k, v in arg_dict.items() if k in field_set}
        return class_name(**filtered)

    def read_expression_data(self, ingest_id: str) -> DataFrame:
        """Get expression data from database.

        Reads full data for a given ingest_id, inefficient and slow.
        Do not use, too slow, use search and
        """
        url = "{}/{}".format(self._get_bulk_url(), ingest_id)

        with requests.Session() as s:
            download = s.get(url)
            decoded_content = download.content.decode("utf-8")
            frame: DataFrame = pandas.read_csv(io.StringIO(decoded_content))
            return frame

    def sort_by_field(
        self, prop: str, expressions: List[DataResult]
    ) -> Mapping[str, List[DataResult]]:
        """Sort the data results by any of their properties.

        @param property: String e.g. "strain" to sort by strain
        @param the raw list of data results returned from a 'search' call.
        """
        ret = {}
        for dr in expressions:
            pvalue = getattr(dr, prop)

            collection = ret.get(pvalue, None)
            if collection is None:
                collection = []
                ret[pvalue] = collection
            collection.append(dr)

        return ret

    def frame(
        self, data: Mapping[str, StrainResult], strain: str, indiv_name: str, sex: Sex
    ) -> DataFrame:
        """Convert a dictionary of gene expression to frame."""
        res: StrainResult = data[strain]

        ids: List[str] = res.gene_ids
        exprs: List[float] = res.strain_expressions[
            "{}@{}".format(indiv_name, sex.name)
        ]
        ret: DataFrame = pandas.DataFrame(
            {"gene_id": numpy.array(ids), "expr": numpy.array(exprs)}
        )
        return ret

    def random(self, ingest_id: str, size: int, count: int = 1) -> List[DataFrame]:
        """Get a random gene expression frame.

        @param ingest_id: from which we ingested data
        @param size: size of the geneset
        """
        url = "{}/{}?gsize={}&random_size={}".format(
            self._get_random_url(), ingest_id, size, count
        )
        response = self._get(url)

        csv_data: List[str] = list(response.text.split("\n"))

        # We return them as one long array which should be faster on the BQ side.
        # Then we split them into sections of size count
        ret: List[List[str]] = self._split_list(csv_data, size)

        return [self._frame(r) for r in ret]

    def random_spearmanrho(
        self, ingest_id: str, scores: List[float], r_size: int = 1, timeout: int = 3600
    ) -> List[float]:
        """Get a random gene expression frame and process random rhos.

        @param ingest_id: from which we ingested data
        @param size: size of the geneset
        """
        url = self._get_random_spearmanrho_url()
        nvr: NullVarianceRequest = NullVarianceRequest(
            id=ingest_id, scores=list(scores), rSz=r_size
        )
        response = self._post(url, nvr.__dict__, timeout=timeout)

        rhos: List[float] = response.json()

        return rhos

    def _split_list(self, lst: List, chunk_size: int) -> List[List]:
        return list(zip(*[iter(lst)] * chunk_size))

    def _frame(self, randoms: List[str]) -> DataFrame:
        # Make them into a frame.
        csv_content = "\n".join(randoms) + "\n"
        csv_content = "{}{}".format("indiv_name,score\n", csv_content)
        ret: DataFrame = pandas.read_csv(StringIO(csv_content))
        return ret

    def _get_search_url(self) -> str:
        return "{}{}".format(self.url, "/gene/expression/search")

    def _get_search_expression_url(self) -> str:
        return "{}{}".format(self.url, "/gene/expression/search-expression")

    def _get_distinct_url(self) -> str:
        return "{}{}".format(self.url, "/meta/distinct")

    def _get_meta_url(self) -> str:
        return "{}{}".format(self.url, "/meta/where/tissue/is")

    def _get_bulk_url(self) -> str:
        return "{}{}".format(self.url, "/bulk/all/where/ingest/is")

    def _get_random_url(self) -> str:
        return "{}{}".format(self.url, "/bulk/random/where/ingest/is")

    def _get_random_spearmanrho_url(self) -> str:
        return "{}{}".format(self.url, "/bulk/random/spearmanrho/")

    def _post(self, url: str, postable_object: dict, timeout: int = 3600) -> Response:

        cookies = None
        if self.auth_proxy is not None:
            cookies = {"_oauth2_proxy": self.auth_proxy}

        with requests.Session() as s:
            response = s.post(
                url, None, postable_object, cookies=cookies, timeout=timeout
            )

            if not response.ok:
                response.raise_for_status()

            return response

    def _get(self, url: str) -> Response:

        cookies = None
        if self.auth_proxy is not None:
            cookies = {"_oauth2_proxy": self.auth_proxy}

        with requests.Session() as s:
            response = s.get(url, cookies=cookies)

            if not response.ok:
                response.raise_for_status()

            return response

    def read_scores(self, path: str) -> Mapping[str, str]:
        """Will read first two columns of csv file.

        into a dictionary of gene: log2fc for use in concordance calc.
        @param path: Path to read.
        """
        # We keep the keys in order here as
        # it is easier to use the debugger and
        # check the dict
        gene_values = OrderedDict()
        with open(path, "r") as file:
            # reading each line from original text file
            for line in file.readlines():
                line = line.strip()
                if not (line.startswith("#")):
                    sa = line.split(",")
                    # Ingore other columns, just use first two.
                    gene_values[sa[0]] = sa[1]

        return gene_values



"""Geneweaver API CLI for Gene endpoints."""

import json
from typing import List, Optional

import typer
from geneweaver.client.api import genes
from geneweaver.client.auth import get_access_token

cli = typer.Typer()

HELP_MESSAGE = """
Tools and utilities for interacting with the Geneweaver API.
"""


@cli.command()
def map_ids(
    ctx: typer.Context,
    source_ids: List[str],
    target_id_type: genes.GeneIdentifier,
    source_id_type: Optional[genes.GeneIdentifier] = None,
    target_species: Optional[genes.Species] = None,
    source_species: Optional[genes.Species] = None,
) -> dict:
    """Map homologs between species."""
    result = genes.map_homologs(
        get_access_token(),
        source_ids=source_ids,
        target_id_type=target_id_type,
        source_id_type=source_id_type,
        target_species=target_species,
        source_species=source_species,
    )

    if len(result["gene_ids_map"]) > 0:
        if not ctx.obj["quiet"]:
            if ctx.obj["pretty"]:
                typer.echo(json.dumps(result, indent=4))
            else:
                typer.echo(json.dumps(result))

    return result



"""Module containing the CLI for the datasets subcommand."""

import typer

cli = typer.Typer()


@cli.command(name="list")
def _list() -> None:
    """List all available datasets."""
    print("- DNACombinedaCGHGeneSummary")


@cli.command()
def downloaded() -> None:
    """List downloaded datasets."""
    print("- DNACombinedaCGHGeneSummary")



"""Utility functions for use with the CLI."""

import functools
from typing import Callable

import typer


def print_value_errors(func: Callable) -> Callable:
    """Print any value errors that occur in the decorated function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):  # noqa: ANN002, ANN003, ANN202
        try:
            return func(*args, **kwargs)
        except ValueError as e:
            print(f"An error occurred: {e}")
            typer.Exit(code=1)

    return wrapper



"""Utility functions for prompting for generic values."""

from typing import Optional, Type, TypeVar

import typer
from geneweaver.client.utils.cli.prompt.none import (
    allow_none_str,
    prompt_if_none,
    value_represents_none,
)

T = TypeVar("T")


def prompt_generic(field_name: str, field_type: Type[T], allow_none: bool = False) -> T:
    """Prompt the user for a value for a field."""
    text_prompt = f"{field_name.replace('_', ' ').capitalize()} ({field_type.__name__})"
    if allow_none:
        text_prompt += allow_none_str()

    while True:
        result = typer.prompt(text_prompt)

        if allow_none and value_represents_none(result):
            return None
        else:
            try:
                return field_type(result)
            except (TypeError, ValueError):
                print(f"Invalid value for {field_name}. Please try again.")
                continue


def prompt_to_keep_dict_field(
    field_name: str, existing_kwargs: dict, default: bool = True
) -> bool:
    """Prompt the user to keep an existing field value.

    :param field_name: The name of the field to prompt for.
    :param existing_kwargs: The existing kwargs to use as defaults.
    :param default: What to set the default value to.
    :return: True if the user wants to keep the field value, False otherwise.
    """
    if field_name in existing_kwargs:
        selection = prompt_to_keep_field(
            field_name, existing_kwargs[field_name], default=default
        )
        if not selection:
            existing_kwargs.pop(field_name)
        return selection
    return False


def prompt_to_keep_field(
    field_name: str, field_value: str, default: bool = True
) -> bool:
    """Prompt the user to keep a field value.

    :param field_name: The name of the field to prompt for.
    :param field_value: The value of the field to prompt for.
    :param default: What to set the default value to.
    :return: True if the user wants to keep the field value, False otherwise.
    """
    field_name = field_name.replace("_", " ").capitalize()
    return typer.confirm(f"{field_name} = {field_value}, OK?", default=default)


def prompt_if_none_or_ask_to_keep(
    field_name: str, value: Optional[T] = None, default: bool = True
) -> Optional[T]:
    """Prompt the user to enter a value if the value is None or to keep the value.

    :param field_name: The name of the field to prompt for.
    :param value: The value to check.
    :param default: The default value for the prompt.
    :return: The value entered by the user.
    """
    if value is None:
        value = prompt_if_none(field_name, value)
    elif not prompt_to_keep_field(field_name, value, default=default):
        value = prompt_if_none(field_name, None)
    return value
